// ignore_for_file: unnecessary_null_comparison, null_argument_to_non_null_type

import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:crypto/crypto.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:package_info/package_info.dart';
import 'package:path_provider/path_provider.dart';

import 'package:vbdler/vbdler.dart';

class VirtualAssetBundle extends CachingAssetBundle {
  bool _inited = false;
  File? _file;
  late VBDLHeader? _header;
  late Map<String, List<int>>? _indexMap;

  Future<Uint8List> _openRead(int location, int length) async {
    final stream = _file?.openRead(location, location + length);

    final Completer<Uint8List> completer = Completer<Uint8List>.sync();
    final List<List<int>> chunks = <List<int>>[];
    int contentLength = 0;
    stream?.listen((List<int> chunk) {
      chunks.add(chunk);
      contentLength += chunk.length;
    }, onDone: () {
      final Uint8List bytes = Uint8List(contentLength);
      int offset = 0;
      for (List<int> chunk in chunks) {
        bytes.setRange(offset, offset + chunk.length, chunk);
        offset += chunk.length;
      }
      completer.complete(bytes);
    }, onError: completer.completeError, cancelOnError: true);

    return completer.future;
  }

  Future<List<int>> _lookup(String key) async {
    if (_inited == false) await _init();
    if (_file == null || _header == null || _indexMap == null)
      return Future.value(null);
    return Future.value(_indexMap?[key]);
  }

  Uint8List _decryptData(Uint8List bytes, int random) {
    return Uint8List.fromList(decryptData(bytes, random));
  }

  Future<String> _resolvePath() async {
    final tmpDir = await getTemporaryDirectory();
    final assetsVD = Directory([tmpDir.path, "VBDL"].join("/"));
    assetsVD.createSync(recursive: true);
    
    // Reset File
    File vbdlF = File([assetsVD.path, VBDL_ASSETS_KEY].join("/"));
    if (vbdlF.existsSync()) vbdlF.deleteSync();
    try {
        final assetsData = await rootBundle.load(VBDL_ASSETS_KEY);
        vbdlF = await vbdlF.writeAsBytes(Uint8List.view(assetsData.buffer), mode: FileMode.writeOnly, flush: true);
    } catch(e) {
        return Future.value(null);
    }
    if (vbdlF.existsSync()) return Future.value(vbdlF.path);
    else return Future.value(null);
  }

  Future<String> _resolveKey() async {
    final info = await PackageInfo.fromPlatform();
    final bdlId = info.packageName;
    final rawKey = "$VBDL_HEADER_ENCKEY.$bdlId.$VBDL_DATA_ENCKEY";
    final key = md5.convert(utf8.encode(rawKey)).toString();
    return Future.value(key);
  }

  Future<bool> _init([String? key, String? path]) async {
    if (_inited == true) return Future.value(false);
    _inited = true;

    if (key == null) key = await _resolveKey();
    if (key == null) return Future.value(false);

    if (path == null) path = await _resolvePath();
    if (path == null) return Future.value(false);

    _file = File(path);
    if (_file?.existsSync() != true) return Future.value(false);

    try {
      bindIndexDecKey(key);

      final headerData = await _openRead(0, VBDLHeader.ENCRYPTED_LEN);
      _header = decryptHeader(headerData);

      final indexData = await _openRead(_header!.indexStart, _header!.indexLength);
      _indexMap = decryptIndex(indexData);

      return Future.value(true);
    } catch(e) {
      _header = null;
      _indexMap = null;
      return Future.value(false);
    }
  }

  @override
  Future<ByteData> load(String key) async {
    // AssetManifest.json

    ByteData? asset;
    final foundDesc = await _lookup(key);
    if (foundDesc != null) {
      assert(foundDesc.length == 3);

      final dataOffset = foundDesc[0];
      final dataSize = foundDesc[1];
      final dataRandom = foundDesc[2];

      try {
        final encBytes =
            await _openRead(_header!.dataStart + dataOffset, dataSize);
        final decBytes = _decryptData(encBytes, dataRandom);
        asset = ByteData.view(decBytes.buffer);
      } catch (e) {
        asset = null;
      }
    } else {
      // 如果在VBDL中没有找到,则在rootBundle中查找
      try {
        asset = await rootBundle.load(key);
      } catch (e) {
        asset = null;
      }
    }

    if (asset == null) throw FlutterError('Unable to load asset: $key');
    return asset;
  }

  Future<void> loadFont(String family, String key) async {
    final fontLoader = FontLoader(family);
    fontLoader.addFont(load(key));
    return fontLoader.load();
  }
}

// VBDL
final AssetBundle virtualBundle = VirtualAssetBundle();
